# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

try:
    from ._models_py3 import AclFailedEntry
    from ._models_py3 import FileSystem
    from ._models_py3 import FileSystemList
    from ._models_py3 import LeaseAccessConditions
    from ._models_py3 import ModifiedAccessConditions
    from ._models_py3 import Path
    from ._models_py3 import PathHTTPHeaders
    from ._models_py3 import PathList
    from ._models_py3 import SetAccessControlRecursiveResponse
    from ._models_py3 import SourceModifiedAccessConditions
    from ._models_py3 import StorageError
    from ._models_py3 import StorageErrorAutoGenerated
except (SyntaxError, ImportError):
    from ._models import AclFailedEntry  # type: ignore
    from ._models import FileSystem  # type: ignore
    from ._models import FileSystemList  # type: ignore
    from ._models import LeaseAccessConditions  # type: ignore
    from ._models import ModifiedAccessConditions  # type: ignore
    from ._models import Path  # type: ignore
    from ._models import PathHTTPHeaders  # type: ignore
    from ._models import PathList  # type: ignore
    from ._models import SetAccessControlRecursiveResponse  # type: ignore
    from ._models import SourceModifiedAccessConditions  # type: ignore
    from ._models import StorageError  # type: ignore
    from ._models import StorageErrorAutoGenerated  # type: ignore

from ._azure_data_lake_storage_restapi_enums import (
    PathExpiryOptions,
    PathGetPropertiesAction,
    PathLeaseAction,
    PathRenameMode,
    PathResourceType,
    PathSetAccessControlRecursiveMode,
    PathUpdateAction,
)

__all__ = [
    'AclFailedEntry',
    'FileSystem',
    'FileSystemList',
    'LeaseAccessConditions',
    'ModifiedAccessConditions',
    'Path',
    'PathHTTPHeaders',
    'PathList',
    'SetAccessControlRecursiveResponse',
    'SourceModifiedAccessConditions',
    'StorageError',
    'StorageErrorAutoGenerated',
    'PathExpiryOptions',
    'PathGetPropertiesAction',
    'PathLeaseAction',
    'PathRenameMode',
    'PathResourceType',
    'PathSetAccessControlRecursiveMode',
    'PathUpdateAction',
]
